#------------------------------------------------------------------
# $Copyright: Copyright (c) 2022 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  #use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/PgSQLRep/PgSQLRep.pm,v 1.11 2015/02/19 07:27:11 mrampura Exp $
#
# $sAuthor: $
#
# $Date: 2015/02/19 07:27:11 $
# $Revision: 1.11 $
#
# Provides:                            PgSQL
# Author:                              mrampura
# crAgent Build:                       1.90 2009/06/03 07:07:19
# crAgent Run Date:                    20 Nov, 2009
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;
#
#----------------------
# Declare the package..
#----------------------
package PgSQLRep;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::Basename;
use File::stat;
use threads;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHagrp = $HAGRP ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHasys = $HASYS ||= "$CLUS_HOME/bin/hasys";
my $sCmdSSH = "/usr/bin/ssh";
my $sCmdCP = "/usr/bin/cp";
my $sCmdTouch = "/usr/bin/touch";
my %hsPgSQLAttr = ();
#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: crAgent,v 1.99 2012/07/13 07:51:52 mrampura Exp \$::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define PgSQLRep subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  my ($sPgSQLResource) = @_;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20001, "Subroutine <%s> called with
  PgSQLResource            [%s]",
  $sMyName, $sPgSQLResource);
  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( $sPgSQLResource !~ /\S/o ) {
    &logError ( 20002,
    "<%s> The PgSQLResource attribute is not set.",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  if ( ! PgSQLRep::GetPgSQLAttr($sPgSQLResource, \%hsPgSQLAttr) ) {
    &logError ( 20003, "<%s> Could not fetch PgSQL attributes. Returning failure.",$sMyName);
    $iReturnCode = 0;
    }
    return $iReturnCode ;
  }


=pod

 -------------------------------------------------------------------------
 Function : bool Start ( );

 Description : Start logic needed for online entry pt

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20004, "Subroutine <%s> called.", $sMyName);
  
  my $iTillWhen = $^T + PgSQLRep::GetEngineTimeout ( 1 );
  my $bPromotionFlag = 0;
  my $bIsPrimary = 0;
  my $bRepStatus = 0;
  my $bIsPrimaryRunning = 0;
  my $iret = 0;
  my $bStandbyRepStatus = 0;
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ($sCommand, $sCommandArgs, $iReturnCode, $iExitCode, @lsCmdout, $iTimeout) = (); 
  #---------------------------------------------
  # Check if Database is already Primary
  #---------------------------------------------
  ($iret, $bIsPrimary) = IsPrimary($iTillWhen);
  if (!$iret) {
    &logError ( 20005, "<%s> Could not fetch the PostgreSQL server mode. Returning failure",$sMyName);
    return 0;
  }
  if ( $bIsPrimary ) {
    &logTrace ( 20006, "<%s> This node is already in Primary mode.",$sMyName);
    ($iret, $bRepStatus) = CheckReplicationStatus($iTillWhen);
    if (!$iret) {
      &logError ( 20007, "<%s> Could not obtain the replication status. Returning failure",$sMyName);
      return 0;
    }
    if ( $bRepStatus){
      &logInfo ( 20008, "<%s> Replication Link up.", $sMyName);
    }
    else {
      &logWarn ( 20009, "<%s> Replication Link down.", $sMyName);
    }
  }
  else {
    &logTrace ( 20010, "<%s> This node is in Standby mode.",$sMyName);
    ($iret, $bIsPrimaryRunning ) = IsPrimaryDBReachable ($iTillWhen);
    if (!$iret) {
        &logError ( 20011, "<%s> Could not verify if Primary is reachable. Returning failure",$sMyName);
        return 0;
    }
    elsif ($bIsPrimaryRunning ) {
      &logTrace ( 20012, "<%s> Primary Database is running.", $sMyName);
      ($iret, $bStandbyRepStatus ) = IsReplicationActive ($iTillWhen);
      if (!$iret) {
        &logError ( 20013, "<%s> Could not obtain replication status. Returning failure",$sMyName);
        return 0; 
      }
      elsif ($bStandbyRepStatus){
        &logTrace ( 20014, "<%s> Replication is set. Now the agent will attempt to make promote the standby.",$sMyName);
        $bPromotionFlag = 1;
      }
      else {
        if ($hsPgSQLAttr{'SplitTakeOver'}) {
          &logInfo ( 20015, "<%s> Attribute SplitTakeOver is set. The agent will continue with the promotion procedure.", $sMyName);
          $bPromotionFlag = 1;
        } else {
          &logError ( 20016, "<%s> Attribute SplitTakeOver is not set. The agent will exit now.", $sMyName);
          return 0;
        }
      }
    }
    else {
      if ($hsPgSQLAttr{'AutoTakeOver'}) {
          &logInfo ( 20017, "<%s> Attribute AutoTakeOver is set. The agent will continue with the promotion procedure.", $sMyName);
          $bPromotionFlag = 1;
      } else {
          &logError ( 20018, "<%s> Attribute AutoTakeOver is not set. The agent will exit now.", $sMyName);
          return 0;
      }
    }
  }
  #------------------------------------------
  # PromotionFlag is only set if the database 
  # was in Standby mode when online was
  # issued.
  #------------------------------------------
  if ($bPromotionFlag == 1 ) {
    if (PromoteToPrimary($iTillWhen)) {
      my $sReplicationModeTuples = $hsPgSQLAttr{'ReplicationModeTuples'};
      if ($sReplicationModeTuples ne "") {
        $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
        my ( $sHostLst, @RepModes, $iIdx, $iTotal, $ilstIdx, $flag, $sLast, $sCmdOpts ) = ();
        $sHostLst = "";
        @RepModes = split(' ', $sReplicationModeTuples);
        $iIdx = 1;
        $iTotal = 0;
        $ilstIdx = $#RepModes;
        $flag = 0;
        $sCmdOpts ="";
        while ($iIdx <= $ilstIdx) {
          if ($RepModes[$iIdx] eq "sync") {
            if ($flag) {
              $sLast = ", ";
            } else {
              $sLast = "";
              $flag = 1;
            }
            $sHostLst = $sHostLst . $sLast . "\\\"$RepModes[$iIdx - 1]\\\"";
            $iTotal = $iTotal + 1;
          }
            $iIdx = $iIdx + 2;
        }
        $iTimeout= $iTillWhen - time ;
        if($iTotal){
          $sCmdOpts = "FIRST $iTotal \($sHostLst\)";
        }
        $sCommandArgs = "\"ALTER SYSTEM SET synchronous_standby_names";
        $sCommandArgs = "-A -x -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c $sCommandArgs = \'\"\'\"\'$sCmdOpts\'\"\'\"\'\"";
        if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){
          &logTrace ( 20259,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
       	  $iTimeout= $iTillWhen - time ;
          $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"SELECT pg_reload_conf();\"";
          &logInfo ( 20260, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );

	        if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){
          	  &logTrace ( 20261,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
		
          } else {
            &logError ( 20262,
                      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
                      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
            return 0; # Failure..
          }
        } else {
          &logError ( 20263,
                      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
                      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
          return 0; # Failure..
        }
      } 

      &logTrace ( 20019, "<%s> Promotion successful.", $sMyName);
    } else {
      # Could not promote to primary. return failure.
      &logError ( 20020, "<%s> Promotion failed. The agent will exit now.", $sMyName);
    return 0;
    }#PromoteToPrimary
  }

  if (!$bIsPrimary ) {                                     # RegistrationOfStandby not attempted if already primary(r/w)
    &logTrace ( 20021, "<%s> The agent will attempt to shut down the old primary.", $sMyName);
    if (ShutDownAndRegisterStandby($iTillWhen)){
      &logTrace ( 20022, "<%s> Old primary  or few standbys are not running.", $sMyName);
      if ($hsPgSQLAttr{'RegistrationOfStandby'} == 1) {
      ($iret, $bRepStatus) = CheckReplicationStatus($iTillWhen);
      if (!$iret) {
        &logError ( 20023, "<%s> Could not obtain the replication status. Returning failure",$sMyName);
        return 0;
      }
      elsif ($bRepStatus){
        &logTrace ( 20024, "<%s> Replication Link up. Returning success.", $sMyName);
      }
      else {
        &logWarn ( 20025, "<%s> Replication Link down. The agent will continue based on SplitTakeOver value", $sMyName);
        if ($hsPgSQLAttr{'SplitTakeOver'} == 1) {
          &logInfo ( 20026, "<%s> SplitTakeOver set. Returning success.", $sMyName);
        }
        else {
          &logError ( 20027, "<%s> SplitTakeOver not set. Returning failure.", $sMyName);
          return 0;
        }  
      }
    }
   }
    else {
      &logWarn ( 20028, "<%s> Action at Standby node failed.", $sMyName);
      if ($hsPgSQLAttr{'SplitTakeOver'} == 1) {
        &logInfo ( 20029, "<%s> SplitTakeOver set. Returning success.", $sMyName);
      }
      else {
        &logError ( 20030, "<%s> SplitTakeOver not set. Returning failure.", $sMyName);
        return 0;
      }
    }
  }
  else {
    &logTrace ( 20031, "<%s> Either old primary not found or I am primary db.", $sMyName);
  } 
  if ((!$bIsPrimary)         || 
      (!$hsPgSQLAttr{'LinkMonitor'} || $bRepStatus)) {
    if ( CreateLockFile () ){
      &logTrace ( 20032, "<%s> Successfully created the lock file.", $sMyName);
    }else  {
      # Could not create lock file.  return failure.
      &logError ( 20033, "<%s> Could not create lock file. The agent will exit now.", $sMyName);
      return 0;
    }#CreateLockFile
  }
  # return success
   return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop (  );

 Description : Stop logic needed for offline entry pt

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20034, "Subroutine <%s> called",
  $sMyName);
  my $iReturnCode = 1; # Success..
  if ($iReturnCode = RemoveLockFile()){
    &logTrace ( 20035, "<%s>: Successfully removed the lock file.",
         $sMyName);
  }
  else {
    &logError ( 20036, "<%s>: Failed to remove the lock file. Returning failure.",
       $sMyName);
    $iReturnCode = 0;
  }
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {

  #----------------
  # Get Arguments..
  #----------------
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20037, "Subroutine <%s> called",
  $sMyName);
  my $iReturnCode = 1; # Success..
  if ($iReturnCode = RemoveLockFile()){
    &logTrace ( 20038, "<%s>: Successfully removed the lock file.",
         $sMyName);
  }
  else {
    &logError ( 20039, "<%s>: Failed to remove the lock file. Returning failure.",
       $sMyName);
    $iReturnCode = 0;
  }
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( );

 Description : First level monitor to determine state of resource.

 Parameters :

 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my $sMyName = ( caller 0 )[3];
  my ( $iWalSenderPid ) = @_;
  &logTrace ( 20040, "Subroutine <%s> called ", $sMyName);
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my %hspg_stat_rep = ();
  my ($iret, $bSync, $bRepStatus, @sReplicationModeArr, $sSyncSlave, $iIdx, $iIsLast, $iAllDown, $iExitCode, $sWalSenderPids, $sCommand ) = ();
  my $iTillWhen = $^T + PgSQLRep::GetEngineTimeout ( 1 );
  my $iFound = VCSDef_MonitorExit::OFFLINE;
  if (ExistsLockFile()) {
          &logTrace ( 20041, "<%s> LockFile exists.",$sMyName);
          $iFound = VCSDef_MonitorExit::ONLINE;
  } else {
          &logTrace ( 20042, "<%s> LockFile does not exist.",$sMyName); 
         $iFound = VCSDef_MonitorExit::OFFLINE;
         return $iFound;
  }

  @sReplicationModeArr = split( " ", $hsPgSQLAttr{'ReplicationModeTuples'} );
  $sSyncSlave = "";
  $iIdx = 0;
  $iIsLast = 0;
  $iAllDown = 1;
  if(!@sReplicationModeArr){
    $sReplicationModeArr[0] ='sync';
    $iIdx = 1;
  }
  foreach my $sItem ( @sReplicationModeArr ){
    if(!$sReplicationModeArr[$iIdx + 1]) {
      $iIsLast = 1;
    }
    if( $iIdx % 2 != 0 && $sItem eq 'sync' ){
      if ($hsPgSQLAttr{'SwitchMode'} == 1) {
        %hspg_stat_rep = ();
        ($iret, $bRepStatus) = CheckReplicationStatus($iTillWhen, $sSyncSlave, \%hspg_stat_rep);
        if (!$iret) {
          &logError ( 20224, "<%s> Could not obtain the replication status. Returning unknown.",$sMyName);
          return 99;
        }
          &logTrace ( 20225, "<%s> Replication check successful.",$sMyName);
          if (defined $hspg_stat_rep{pid}  &&
            $hspg_stat_rep{pid} =~ /\S/ ) {
            &logTrace ( 20167, "<%s> Standby entry found with PostgreSQL wal sender Pid [%s] for Node [%s].",
                    $sMyName, $hspg_stat_rep{pid}, $sSyncSlave);
            $iAllDown = 0;
            if( $sSyncSlave ){
              $iWalSenderPid = "";
              if ( GetAttrValue("WalSenderPids", $sResourceName, "", \$sWalSenderPids ) ) {
                &logTrace ( 20264, "<%s> Attribute value for Attribute WalSenderPids is [%s]",
                        $sMyName, $sWalSenderPids );
              }
              else {
                &logError ( 20265, "<%s> Failed to obtain attribute value for Attribute WalSenderPids. Returning failure",
                          $sMyName );
              }
              ($iWalSenderPid) = $sWalSenderPids =~ /$sSyncSlave=(\S+)/;
              if ( !$iWalSenderPid ){
                $sWalSenderPids .= " $sSyncSlave=00";
                $iWalSenderPid = "00";
              }
            }
            &logTrace ( 20264, "<%s> WalSenderPid Attribute value for Slave [%s] is [%s]",
                      $sMyName, $sSyncSlave, $iWalSenderPid );
            if ($hspg_stat_rep{pid} != $iWalSenderPid ) {
              &logInfo ( 20168, "<%s> WalSenderPid attribute value does not match the actual Pid value. Setting it to [%s]",
                $sMyName, $hspg_stat_rep{pid});
              if( $sSyncSlave ){
                $sWalSenderPids =~ s/$sSyncSlave=\S+/$sSyncSlave=$hspg_stat_rep{pid}/;
                $sCommand = "$sCmdHares -modify $sResourceName WalSenderPids \"$sWalSenderPids\" ";
              }
              else {
                $sCommand = "$sCmdHares -modify $sResourceName WalSenderPid $hspg_stat_rep{pid} ";
              }
              &logTrace ( 20169, "Running command [%s]", $sCommand );
              readpipe ("$sCommand");
              $iExitCode = $?>>8;
              &logTrace ( 20170, "<%s> ExitCode [%s]", $sMyName, $iExitCode);
              if ( $iExitCode ){
                &logTrace ( 20266, "<%s> Failed to execute command [%s] with ExitCode [%s], Returning failure", $sMyName, $sCommand, $iExitCode);
              }
            }
            # check if state is async
            if ((defined $hspg_stat_rep{sync_state}) &&
                ($hspg_stat_rep{sync_state} eq "async" ) &&
                ($hspg_stat_rep{state} eq "streaming")) {
              # change mode to sync
              &logInfo ( 20171, "<%s> Wal sender Pid [%s], Sync State [%s] state [%s] The agent will now attempt to Switch mode to Sync",
                    $sMyName, $hspg_stat_rep{pid}, $hspg_stat_rep{sync_state}, $hspg_stat_rep{state} );
              if (SwitchToSyncMode($iTillWhen, $sSyncSlave)) {
                # Successfully switched mode to Sync
                &logInfo ( 20172, "<%s> Successfully switched mode from Async to Sync.",
                    $sMyName );
              }
              else {
                # Could not switch mode to sync. return unknown    
                &logWarn ( 20173, "<%s> The agent could not switch mode from Async to Sync.",
                    $sMyName );
                return 99;
              }
            }
            else {
              # Correct mode sync
              &logTrace ( 20174, "<%s> Sync State [%s] state [%s]. The agent will not take any action",
                    $sMyName, $hspg_stat_rep{sync_state}, $hspg_stat_rep{state} );
            }
          }
          else { # Standby down
            # check state
            &logTrace ( 20175, "<%s> No Standby entry found in pg_stat_replication.",
                      $sMyName );
            ($iret, $bSync) = IsSynchronous($iTillWhen);
            if (!$iret) {
              # check sync commit command failed. return unknown
              &logWarn ( 20176, "<%s> The agent could not check synchronous state. Returning unknown.",
                        $sMyName );
              return 99;
            }
            if ($iIsLast && $iAllDown){
              $sSyncSlave ="";
            }
            if ($bSync) {
              # change mode to async
              &logTrace ( 20177, "<%s> Synchronous commit is set to ON. The agent will now attempt to Switch mode to Async",
                    $sMyName );
              if (SwitchToAsyncMode($iTillWhen, $sSyncSlave)) {
                # Successfully switched mode to Async
                &logInfo ( 20178, "<%s> SyncToAsync operation successfull for slave [%s].",
                $sMyName, $sSyncSlave );
              }
              else {
                # Could not switch mode to Async. return unknown        
                &logWarn ( 20179, "<%s> The agent did not switch mode from Sync to Async for slave [%s].",
                    $sMyName, $sSyncSlave );
                return 99;
              }
            }
          }
      } elsif ($hsPgSQLAttr{'DetailedMonitoring'} == 1) {
        ($iret, $bRepStatus) = CheckReplicationStatus($iTillWhen);
        if (!$iret) {
          &logError ( 20226, "<%s> Could not obtain the replication status. Returning unknown",$sMyName);
          return 99;
        }
        if ( $bRepStatus){
          &logTrace ( 20043, "<%s> Detailed Monitoring successful.",$sMyName);
        }
        else {
          &logWarn ( 20044, "<%s> Detailed Monitoring failed.",$sMyName);
        }
      }
    }
    else{
      $sSyncSlave = $sItem;
    }
    $iIdx++;
  }
return $iFound;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sOutputFilename,
       $sHideString ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20045, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20046,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20047,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20048,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20049,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           undef,
                                                           $sHideString);

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20050,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20051,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20052,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20053,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}




=pod

 -------------------------------------------------------------------------
 Function : bool CheckReplicationStatus ( IN int TillWhen);

 Description : Check the replication status

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub CheckReplicationStatus {
  my ($iTillWhen, $sSyncSlave, $rhspg_stat_rep ) = @_;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20054, "Subroutine <%s> called with 
  TillWhen               [%s]
  pg_stat_rep            [%s]",
  $sMyName, $iTillWhen, $rhspg_stat_rep); 
  my (@lsCmdout, $iExitCode, $iReturnCode) = ();
  my $iTimeout= $iTillWhen - time ;
  my $sWhere ="";  
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  if( $sSyncSlave ){
    $sWhere = "where application_name=\'\"\'\"\'$sSyncSlave\'\"\'\"\'";
  }
  my $sCommandArgs = "-A -x -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"select * from pg_stat_replication $sWhere\""; 
  if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){

    &logTrace ( 20055,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
      if ( 0 == $iExitCode ){
        if ($hsPgSQLAttr{'DetailedMonitoring'} == 1) {
                &logInfo ( 20181,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );

        } else {
                &logTrace ( 20056,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
        }
      if ( ( $lsCmdout[0]) && ( grep(/streaming/,@lsCmdout) ) ) {
        &logTrace ( 20057,
                "<%s> Replication active",
                $sMyName );
	        return (1, 1) if ! defined $rhspg_stat_rep;
      }
      else {
        &logTrace ( 20058,
                "<%s> Replication NOT active.",
                $sMyName );
        return (1, 0);
      }

    }
    else {
      &logError ( 20059,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, @lsCmdout );
      return 0 # Failure..
    }
  }
  else {
    &logError ( 20060,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  } 
  foreach my $sLine (@lsCmdout) {
    $sLine =~ m/^(\w+)\s*\|\s*(.*?)$/;
    my $sParameter = $1;
    my $sValue = $2;
    $$rhspg_stat_rep{$sParameter} = $sValue;
    &logTrace ( 20182, "Parameter:[%s] Value [%s] ", $sParameter, $sValue);
  }
return (1, 1);
}
  

=pod

 -------------------------------------------------------------------------
 Function : bool IsReplicationActive (IN int TillWhen );

 Description : Check if the current node is standby and replication is 
 in healthy state.

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub IsReplicationActive {
  my ($iTillWhen) = @_ ;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20061, "Subroutine <%s> called with
  TillWhen               [%s]",
  $sMyName, $iTillWhen); 
  my (@lsCmdout, $iExitCode, $iReturnCode) = ();
  my $iTimeout= $iTillWhen - time ;
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  my $sCommandArgs = "-A -x -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"select * from pg_stat_wal_receiver\""; 
  if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){

    &logTrace ( 20062,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
      if ( 0 == $iExitCode ){
                &logTrace ( 20063,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      if ( ( $lsCmdout[0]) && ( grep(/\|streaming/,@lsCmdout) ) ) {
        &logTrace ( 20064,
                "<%s> Replication active",
                $sMyName );
        return (1, 1);
      }
      else {
        &logWarn ( 20065,
                "<%s> Replication NOT active",
                $sMyName);
        return (1, 0);
      }

    }
    else {
      &logError ( 20066,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, @lsCmdout );
      return 0 # Failure..
    }
  }
  else {
    &logError ( 20067,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }  
return 0;
  
}




=pod

 -------------------------------------------------------------------------
 Function : bool ShutDownAndRegisterStandby( );

 Description : Executes an Action EP on a remote machines to start the
 postgres database in standby mode.

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ShutDownAndRegisterStandby {
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20068, "Subroutine <%s> called",
  $sMyName); 
  my $sRes = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sReplicationModeTuples, @RepModes, $iIdx, $iRet, $sSysName,
      $ilstIdx, @out, $sOutline, $sCommand, @lsOut, $sClusName, $iExitCode) = ();
  $sReplicationModeTuples = $hsPgSQLAttr{'ReplicationModeTuples'};
  if ($sReplicationModeTuples eq "") {
    ($sClusName, $sSysName) = FindRemoteClusterSystem();
    return 1 unless (defined $sClusName);

    $sCommand = "$sCmdHares -action $sRes RegisterStandby -actionargs $hsPgSQLAttr{'RegistrationOfStandby'} -sys $sSysName -clus $sClusName";
    if ($sClusName eq "") {
    $sCommand = "$sCmdHares -action $sRes RegisterStandby -actionargs $hsPgSQLAttr{'RegistrationOfStandby'} -sys $sSysName "; 
    }
    &logTrace ( 20069, "Running action command [%s]", $sCommand );
    readpipe ("$sCommand");
    $iExitCode = $?>>8;
    &logTrace ( 20070, "<%s> ExitCode [%s]", $sMyName, $iExitCode);
    return $iExitCode ? 0 : 1 ;
  } else {
    @RepModes = split(' ', $sReplicationModeTuples);
    $iIdx = 0;
    $ilstIdx = $#RepModes;
    $iRet = 1;
    my %ActionCmd = ();
    while ($iIdx <= $ilstIdx) {
      $sCommand = "$sCmdHasys -state";
      @out = readpipe($sCommand );
      $iExitCode = $?>>8;
      &logTrace ( 20267, "<%s> Command [%s] executed with ExitCode [%s]", $sMyName, $sCommand, $iExitCode);
      if ( $iExitCode ){
        return 0;
      }
      shift @out;
      foreach $sOutline (@out) {
        @lsOut = split( '\s+', $sOutline);
        if ($lsOut[0] =~ /:/) {
          ($sClusName, $sSysName) = split(':',$lsOut[0]);
        } else {
          $sClusName = "";
          $sSysName = $lsOut[0];
        }        
        if ($sSysName eq $RepModes[$iIdx]) {
          last;
        }
      }

      $sCommand = "$sCmdHares -action $sRes RegisterStandby -actionargs $hsPgSQLAttr{'RegistrationOfStandby'} -sys $sSysName -clus $sClusName";
      if ($sClusName eq "") {
        $sCommand = "$sCmdHares -action $sRes RegisterStandby -actionargs $hsPgSQLAttr{'RegistrationOfStandby'} -sys $sSysName "; 
      }
      &logTrace ( 20268, "Running action command [%s]", $sCommand );
      $ActionCmd{$sSysName} = $sCommand; 
      $iIdx = $iIdx + 2;
    }
    my %Threads = ();
    for $sSysName (keys %ActionCmd) {
      $Threads{$sSysName} = threads->create('RunCmd', $ActionCmd{$sSysName});
    }
    for $sSysName (keys %ActionCmd) {
        $iExitCode = $Threads{$sSysName}->join();
        if ($iExitCode and $iRet) {
          $iRet = $iExitCode;
        }
    }
  }
  return $iRet;
}

sub RunCmd {
	my ($sCommand) = @_;
  my $sMyName = ( caller 0 )[3];
	readpipe ($sCommand);
  my $iExitCode = $?>>8;
  &logTrace ( 20269, "<%s> Command [%s] executed with ExitCode [%s]", $sMyName, $sCommand, $iExitCode);
  return $iExitCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool PromoteToPrimary ( IN Integer TillWhen );

 Description : Promote standby to primary.

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub PromoteToPrimary {
  my ($iTillWhen) = @_;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20071, "Subroutine <%s> called with
  TillWhen           [%s]",
  $sMyName, $iTillWhen); 
  my (@lsCmdout, $iExitCode, $iReturnCode) = ();
  my $iTimeout= $iTillWhen - time ;
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/pg_ctl";
  my $sCommandArgs = "promote -w -D $hsPgSQLAttr{'DataDir'} -o \"-p $hsPgSQLAttr{'Port'} \" ";
  $iTimeout = $iTillWhen - $^T; 
  &logInfo ( 20072, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20073,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  
    if ( 0 == $iExitCode )
    { 
      &logTrace ( 20074,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    { 
      &logError ( 20075,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
    }
  }
  else
  {
    &logError ( 20076,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
 
  ### Check if database is readwrite
  my ($iret, $bIsPrimary) = IsPrimary($iTillWhen);
  if (!$iret) {
    &logError ( 20146, "<%s> Could not fetch the PostgreSQL server mode. Returning failure",$sMyName);
    return 0;
  }
  if ( $bIsPrimary ) {
    &logTrace ( 20147, "<%s> This node is now in Primary mode.",$sMyName);
  }
  else {
    &logTrace ( 20148, "<%s> Promotion failed. Returning failure. ",$sMyName);
    return 0; 
  }
  
  ### checkpoint
  $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"checkpoint\"";
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20077, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20078,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
      &logTrace ( 20079,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20080,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20081,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : bool IsPrimary ( IN Integer TillWhen );

 Description : Checks if the underlying PostgreSQL DB is running 
 in Primary mode

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 standby
           1 primary
 -------------------------------------------------------------------------

=cut
sub IsPrimary {
  my ($iTillWhen) = @_;
  my $sMyName = ( caller 0 )[3];
  my $Mode = ();
  &logTrace ( 20082, "Subroutine <%s> called with
  TillWhen           [%s]",
  $sMyName, $iTillWhen); 
  my (@lsCmdout, $iExitCode, $iReturnCode) = ();
  my $iTimeout= $iTillWhen - time ;
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  my $sCommandArgs = "-A -x -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"select pg_is_in_recovery\(\)\"";
  
  if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){

    &logTrace ( 20083,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
      if ( 0 == $iExitCode ){
      &logTrace ( 20084,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      chomp $lsCmdout[0];
      if ( grep(/pg_is_in_recovery\|f/,@lsCmdout) ) {
        &logInfo (20085, "This is primary server.");
        return (1,1);
      }
      elsif ( grep(/pg_is_in_recovery\|t/,@lsCmdout) ) {
        &logInfo (20086, "This is Standby server.");
        return (1,0);
      }
    }
    else {
      &logError ( 20087,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, @lsCmdout );
      return 0 # Failure..
    }
  }
  else {
    &logError ( 20088,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }  
return 0;
}

=pod

 -------------------------------------------------------------------------
 Function : bool GetLockFileName ( );

 Description : Gets the lock file name for the resource

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub GetLockFileName {
  my $sMyName = ( caller 0 )[3];
    &logTrace ( 20089, "Subroutine <%s> called",
  $sMyName);
  my $sRes = $ENV { VCS_LOG_RESOURCE_NAME };
  my $lock_file = "$VCSLOG/lock/PgSQLRep/$sRes.lck";

    &logTrace ( 20090, "lockfile is [%s]",
  $lock_file);
    return $lock_file;
}

=pod

 -------------------------------------------------------------------------
 Function : bool CreateLockFile ( );

 Description : Create the lock file.

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub CreateLockFile {
    my ($lock_file) = GetLockFileName();
    my $sMyName = ( caller 0 )[3];
  &logTrace ( 20091, "Subroutine <%s> called",
  $sMyName);
    my (undef, $lock_dir, undef) = fileparse($lock_file);

    if (! -d $lock_dir) {
        if (! mkdir  $lock_dir){
            &logTrace ( 20092, "<%s> Could not create lock directory [%s]",
                    $sMyName, $lock_dir);
           return 0;
        }
    }

    if (!open(LFH, ">$lock_file")){
        &logTrace ( 20093, "<%s> Could not create file: [%s]",
                $sMyName, $lock_file);
        return 0;
    }

    close(LFH);
}


=pod

 -------------------------------------------------------------------------
 Function : bool RemoveLockFile ( );

 Description : Removes the lock file.

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub RemoveLockFile {
    my $sMyName = ( caller 0 )[3];
  &logTrace ( 20094, "Subroutine <%s> called",
  $sMyName);

    my ($lock_file) = GetLockFileName();
    if (-f "$lock_file") {
        if (!unlink("$lock_file")) {
            &logTrace ( 20095, "<%s> Could not delete file: [%s]",
                    $sMyName, $lock_file);
            return 0;
        }
    }
  return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExistsLockFile ( );

 Description : Checks if the lock file exists.

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub ExistsLockFile {
    my $sMyName = ( caller 0 )[3];
  &logTrace ( 20096, "Subroutine <%s> called",
  $sMyName);
    my $retval = 0;
    my ($lock_file) = GetLockFileName();

    if (-f "$lock_file") {
        $retval = 1;
    }
    return $retval;
}


=pod

 -------------------------------------------------------------------------
 Function : bool FindRemoteClusterSystem ( );

 Description : Runs ha commands to find the name of the remote cluster
 where the PgSQLRep global SG is configured.

 Parameters :

 
 Returns : Name of the remote cluster.

 -------------------------------------------------------------------------

=cut
sub FindRemoteClusterSystem {
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20097, "Subroutine <%s> called",
  $sMyName);
  my ($sGroupName, $sCommand, $sCommandArgs )=();
  my $sRes  = $ENV { VCS_LOG_RESOURCE_NAME };
  GetAttrValue("Group", $sRes, "", \$sGroupName);
  $sCommand = "$sCmdHagrp -state $sGroupName";
  &logTrace ( 20098, "Command :[%s]",$sCommand );
  my @out = readpipe($sCommand );
  &logTrace ( 20099, "output[%s]",join(" ", @out ) );
  shift @out;
  my ($sCluster1, $sSystem1, $sOutline, @lsOut, $RemoteCluster, $RemoteSystem, $sCurrentHost) = ();
  GetNodeId (\$sCurrentHost ); 
  foreach $sOutline (@out) {
    @lsOut = split( '\s+', $sOutline);
    if ( $lsOut[2] =~ /:/) {
         ($sCluster1, $sSystem1) = split(':',$lsOut[2]);
    }
    else {
          $sCluster1 = "";
          $sSystem1 = $lsOut[2];
    }
    if ( ($sCluster1 ne "localclus") && ( $sSystem1 ne $sCurrentHost)) {
       $sCommand = $sCmdHasys; # Get global..
       $sCommandArgs = "-state $sSystem1";
       my $sSysState = ();
       ( $sSysState = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                        =~ s/\n//;
       if ( $sSysState eq "RUNNING") {
        $RemoteCluster = $sCluster1;
        $RemoteSystem = $sSystem1;
        last;
      }
    }
  }
  return (undef, undef) unless (defined $RemoteSystem);

  &logTrace ( 20100, "Remote cluster [%s]",$RemoteCluster);
  return ($RemoteCluster, $RemoteSystem);
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetAttrValue ( );

 Description : Runs hares -value command to find the value of the attribute.

 Parameters :

 
 Returns : Attribute value
 -------------------------------------------------------------------------

=cut
sub GetAttrValue{
    my ( $sAttributeName,
         $sResourceName,
         $sSystem,
         $sAttributeValue ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20101, "Subroutine <%s> called with :
  AttributeName              [%s]
  ResourceName               [%s]
  System                     [%s]
  AttributeValue             [%s]",
  $sMyName, $sAttributeName, $sResourceName, $sSystem, $$sAttributeValue );

  my ( @lsHaresOut, @lsOut, $sCommand, $sCommandArgs ) = ();
  my $iExitCode = 0;


  #------------------------------------------------------
  # Sample output:
  # #hares -value PgSQLrep_res Group vcsesx001 -localclus
  #  0
  #-------------------------------------------------------
  $sCommand = $sCmdHares;
  $sCommandArgs = "-value $sResourceName $sAttributeName $sSystem -localclus";
  &logTrace ( 20102, "<%s> Executing Command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
  ( $$sAttributeValue = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                               =~ s/\n//;

  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20103,
                "<%s> Cmd [%s] with args [%s] failed with exit code [%s].",
                $sMyName, $sCommand, $sCommandArgs, $iExitCode);
    return 0;
  }
  else {
    &logTrace ( 20104,
                "<%s> Cmd [%s] with args [%s] succeeded with exit code [%s].",
                $sMyName, $sCommand, $sCommandArgs, $iExitCode);
  }

  if (! (defined $$sAttributeValue) ) {
    &logError ( 20105,
      "<%s> Could not find the [%s] Attribute value.",
      $sMyName, $sAttributeName);
    return 0;
  }
  &logTrace ( 20106,
      "<%s> [%s] Attribute value [%s].",
      $sMyName, $sAttributeName, $$sAttributeValue );

  return 1;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetPgSQLAttr ( IN String PgSQLRes,
                                OUT Hash PgSQLAttr);

 Description : Gets values of all PgSQL attributes

 Parameters :

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub GetPgSQLAttr {
   my ( $sPgSQLRes,
        $rhsPgSQLAttr ) = @_;

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20107, "Subroutine <%s> called with :
  PgSQLRes              [%s]
  PgSQLAttr             [%s]",
  $sMyName, $sPgSQLRes, $rhsPgSQLAttr );
  
  my @lsAttr = qw ( PostgreSQLUser EnvFile HostName Port BaseDir SwitchMode
                    DataDir SourceConnStr RecoveryFile RegistrationOfStandby
                    LinkMonitor AutoTakeOver SplitTakeOver DetailedMonitoring ReplicationModeTuples );
  my $sAttributeName = ();
  foreach $sAttributeName ( @lsAttr ) {
    if (GetAttrValue ($sAttributeName,
                      $sPgSQLRes,
                      "",
                      \$$rhsPgSQLAttr{$sAttributeName} ) ) {

      &logTrace ( 20108, "<%s> Attribute value for Attribute [%s] is [%s]",
               $sMyName, $sAttributeName, $$rhsPgSQLAttr{$sAttributeName} );
    }
    else {
      &logError ( 20109, "<%s> Failed to obtain attribute value for Attribute [%s]. Returning failure",
                $sMyName, $sAttributeName );
      return 0;
    }
  }
return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20110, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20111,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20112,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20113, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20114, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "PgSQLRep" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20115,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20116, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20117, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20118,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : bool IsPrimaryDBReachable ( IN Integer TillWhen );

 Description : Checks if a remote Primary is reachable.

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub IsPrimaryDBReachable {
  my ($iTillWhen) = @_;
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20119, "Subroutine <%s> called with 
  TillWhen               [%s]",
  $sMyName, $iTillWhen ); 
  my (@lsCmdout, $iExitCode, $iReturnCode, $sGroupName, $sLastPrimary) = ();
  my $sSrcConnStr = $hsPgSQLAttr{'SourceConnStr'};
  my $sReplicationModeTuples = $hsPgSQLAttr{'ReplicationModeTuples'};
  if ($sReplicationModeTuples ne "") {
	  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
	  GetAttrValue("Group", $sResourceName, "", \$sGroupName);
	  my $sCmd = "$sCmdHagrp -value $sGroupName UserStrGlobal ";
	  $sLastPrimary = readpipe ( $sCmd ); 
	  chomp($sLastPrimary);
	  &logTrace ( 20233, "Cmd [%s] output [%s] ",
			      $sCmd, $sLastPrimary );
	  if ( $iExitCode = $? >> 8 ) {
	    &logError ( 20234,
		  " Cmd [%s] failed with exit code [%s].",
		  $sCmd, $iExitCode);
	    return (0,0); # Failure..  
	  }
	  $sSrcConnStr =~ s/%Master%/$sLastPrimary/g;
  }
  my $sPrimaryHost = ($sSrcConnStr =~ m/host=(.*?)[\s|\$]/g)? $1 : undef ;
  my $sPrimaryPort = ($sSrcConnStr =~ m/port=(.*?)[\s|\$]/g)? $1 : undef;
  if (!defined $sPrimaryHost) {
    &logError ( 20120,
    "<%s> PrimaryHost [%s] not defined in SourceConnStr[%s]",
    $sMyName,  $sPrimaryHost, $hsPgSQLAttr{'$sSourceConnStr'});
    return (0,0); # Failure..    
  }
   if (!defined $sPrimaryPort){
    $sPrimaryPort = 5432;
  }
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  my $sCommandArgs = "-A -x -h $sPrimaryHost -p $sPrimaryPort -c \"select 1;\"";
  &logTrace ( 20121, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  my $iTimeout = $iTillWhen - time;
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20122,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
                &logTrace ( 20123,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logTrace ( 20124,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
       return (1, 0) # Failure..
    }
  }
  else
  {
    &logError ( 20125,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return (1, 0); # Failure..
  }
  return (1, 1);
}


=pod

--------------------------------------------------------------------------------

 Function : String GetRegex ( IN String Password );

 Description : Prefixes the special characters
 with an extra "\" character.

 Parameters :
 String  Password

 Returns : String where the special characters are prefixed by an extra "\" .

--------------------------------------------------------------------------------

=cut

sub GetRegex
{
  my ($sPassword) = @_;
  $sPassword =~ s/\./\\\./g;
  $sPassword =~ s/\@/\\\@/g;
  $sPassword =~ s/\$/\\\$/g;
  $sPassword =~ s/\(/\\(/g;
  $sPassword =~ s/\)/\\)/g;
  $sPassword =~ s/\^/\\^/g;
  $sPassword =~ s/\[/\\[/g;
  $sPassword =~ s/\]/\\]/g;
  $sPassword =~ s/\{/\\{/g;
  $sPassword =~ s/\}/\\}/g;
  return $sPassword;
}


=pod
--------------------------------------------------------------------------------
 Function: bool GetNodeId( OUT Reference NodeId )

 Description: This subroutine returns VCS nodeid of host where agent is running

 Parameters:
 NodeId            Node name of current host 

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved node name.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub GetNodeId{

  my $sCurrentHost = shift;
  my $sMyName = ( caller 0 )[3];
  &logTrace( 20126, "Subroutine <%s> called:
                     CurrentHost [%s]", $sMyName, $sCurrentHost );
  my $iExitCode = 0;
  my $sCommand = $sCmdHasys;
  my $sCommandArgs = " -nodeid ";
  $$sCurrentHost = readpipe ( "$sCommand $sCommandArgs" );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20127, "<%s>: Cmd [%s] with args [%s] failed with exit code [%s].", $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    return 1;
  }

  chomp $$sCurrentHost;
  $$sCurrentHost =~ s/.*Node name: //;
  $$sCurrentHost =~ s/ Node ID: .*//;
  &logTrace( 20128, "<%s>: Current host is [%s]", $sMyName, $$sCurrentHost );

  return 0;
}

=pod
--------------------------------------------------------------------------------
 Function: bool OpenEP()

 Description: This subroutine handles the LockFile in reboot-like cases

 Parameters:
 NodeId            Node name of current host 

 Returns : boolean value in the set of {0,1} where
           0 Successfully retrieved node name.
           1 Unsuccessful.
--------------------------------------------------------------------------------
=cut

sub OpenEP{
  my $sMyName = ( caller 0 )[3];
  my ($sPgSQLRes) = @_;
  &logTrace( 20149, "Subroutine <%s> called.", $sMyName );
  my ($sCurrentHost, $bPgResOnline, $sPGResState, $iret, $bIsPrimary) = ();
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  if (!ExistsLockFile()) {
          &logTrace ( 20150, "<%s> LockFile does not exist. Open Entry Point will return success.",$sMyName);
         return 1;
  }

  GetNodeId (\$sCurrentHost );
  if (GetAttrValue ("State",
                    $sPgSQLRes,
                    "$sCurrentHost",
                    \$sPGResState ) ) {
      &logTrace ( 20151, "<%s> Attribute value for State is [%s]",
               $sMyName, $sPGResState );
  }
  else {
    &logError ( 20152, "<%s> Failed to obtain attribute value for State. Returning failure",
                $sMyName );
    return 0;
  }

  $bPgResOnline = $sPGResState eq "ONLINE" ? 1 : 0 ;
  if (!$bPgResOnline){
       &logInfo( 20153, "<%s> PostgreSQL resource not online.", $sMyName );
    if (RemoveLockFile()){
      &logTrace ( 20154, "<%s>: Successfully removed the lock file.",
         $sMyName);
      return 1;
    }             
    else {
      &logError ( 20155, "<%s>: Failed to remove the lock file. Returning failure.",
       $sMyName);
      return 0;
    }
  }

  ($iret, $bIsPrimary) = IsPrimary(time + 10);
  if (!$iret) {
    &logError ( 20156, "<%s> Could not fetch the PostgreSQL server mode. Returning failure",$sMyName);
    return 0;
  }
  if ($bIsPrimary) {
      &logInfo ( 20157, "<%s>: PostgreSQL Database is running in primary mode. Returning success.",
         $sMyName);
    return 1;  
  }
  else {
    &logInfo ( 20158, "<%s>: PostgreSQL Database is not running in primary mode. The agent will remove the LockFile.",
         $sMyName);
    if (RemoveLockFile()){
      &logTrace ( 20159, "<%s>: Successfully removed the lock file.",
         $sMyName);
      return 1;
    }
    else { 
      &logError ( 20160, "<%s>: Failed to remove the lock file. Returning failure.",
       $sMyName);
      return 0;
    }
  }
  return 1;
}

=pod

 -------------------------------------------------------------------------
 Function : bool SwitchToAsyncMode ( IN Integer TillWhen );

 Description : Switches the primary database mode from Synchronous to Asynchronus.

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub SwitchToAsyncMode {
  my ($iTillWhen, $sSyncSlave) = @_;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20183, "Subroutine <%s> called with
  TillWhen           [%s]
  SyncSlave          [%s]",
  $sMyName, $iTillWhen, $sSyncSlave); 
  my (@lsCmdout, $iExitCode, $iReturnCode, $sCommand, $sCommandArgs, $sCmdOpts, $iTimeout) = ();
  
  $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  $sCmdOpts = "";
  if ( $sSyncSlave ){
    $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"SHOW synchronous_standby_names\"";
    $iTimeout = $iTillWhen - $^T;
    &logInfo ( 20250, "<%s> Executing Command [%s] with arguments [%s]",
                    $sMyName, $sCommand, $sCommandArgs );
    if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                    $hsPgSQLAttr{'EnvFile'},
                                    $sCommand,
                                    $sCommandArgs,
                                    \$iExitCode,
                                    $iTimeout,
                                    \@lsCmdout ) ) {
      &logTrace ( 20251,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

      if ( 0 == $iExitCode ) {
        &logTrace ( 20252,
                  "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                  $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
        my @sys = $lsCmdout[0] =~ /\((.*)\)/;
        my $iflg = 0;
        foreach my $sNode (split(", ", $sys[0] )){
          if( $sNode =~ $sSyncSlave ){
            $iflg = 1 ;
            &logTrace ( 20253,
            "<%s> Removing Node [%s] from Synchronous_standby_names ",
            $sMyName, $sSyncSlave );
          }
          else{
            chop($sNode);
            $sCmdOpts .= "\\$sNode\\\", ";
          }
        }
        if ($iflg){
          my @sOut =split(" ", $lsCmdout[0]);
          my $iCnt = $sOut[1] - 1;

          if($sCmdOpts){
            $sCmdOpts = "FIRST $iCnt \(".$sCmdOpts;
            chop($sCmdOpts);
            chop($sCmdOpts);
            $sCmdOpts .= "\)";
          }
        }
        else{
          &logTrace ( 20254,
          "<%s> Node [%s] is already not included in Synchronous_standby_names, thus SwitchToAsync Not required, Returning",
          $sMyName, $sSyncSlave );
          return 2;
        }
      }
      else {
        &logError ( 20255,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
         $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
         return 0 # Failure..
      }
    }
    else {
      &logError ( 20256,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      return 0; # Failure..
    }
  }
  else{
    $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"ALTER SYSTEM SET synchronous_commit = off;\"";
    $iTimeout = $iTillWhen - $^T;
    &logInfo ( 20189, "<%s> Executing Command [%s] with arguments [%s]",
                    $sMyName, $sCommand, $sCommandArgs );
    if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                    $hsPgSQLAttr{'EnvFile'},
                                    $sCommand,
                                    $sCommandArgs,
                                    \$iExitCode,
                                    $iTimeout,
                                    \@lsCmdout ) ) {
      &logTrace ( 20190,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

      if ( 0 == $iExitCode )
      {
        &logTrace ( 20191,
                  "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                  $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      }
      else
      {
        &logError ( 20192,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
        $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
        return 0 # Failure..
      }
    }
    else
    {
      &logError ( 20193,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      return 0; # Failure..
    }
  }
  $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"ALTER SYSTEM SET synchronous_standby_names = \'\"\'\"\'$sCmdOpts\'\"\'\"\'\"";
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20184, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20185,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode ) {
      &logTrace ( 20186,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else {
      &logError ( 20187,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else {
    &logError ( 20188,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
  $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"SELECT pg_reload_conf();\"";
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20194, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20195,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
      &logTrace ( 20196,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20197,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20198,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
return 1;
}



=pod

 -------------------------------------------------------------------------
 Function : bool SwitchToSyncMode ( IN Integer TillWhen );

 Description : Switches the primary database mode from Asynchronous to Synchronous.

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut
sub SwitchToSyncMode {
  my ($iTillWhen, $sSyncSlave) = @_;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20199, "Subroutine <%s> called with
  TillWhen           [%s]
  SyncSlave          [%s]",
  $sMyName, $iTillWhen, $sSyncSlave); 
  my (@lsCmdout, $iExitCode, $iReturnCode, $sCommand, $sCommandArgs, $sCmdOpts, $iTimeout) = ();
  
  $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
    
  if ( $sSyncSlave ){
    $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"SHOW synchronous_standby_names\"";
    $iTimeout = $iTillWhen - $^T;
    &logInfo ( 20245, "<%s> Executing Command [%s] with arguments [%s]",
                    $sMyName, $sCommand, $sCommandArgs );
    if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                    $hsPgSQLAttr{'EnvFile'},
                                    $sCommand,
                                    $sCommandArgs,
                                    \$iExitCode,
                                    $iTimeout,
                                    \@lsCmdout ) ) {
      &logTrace ( 20246,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );

      if ( 0 == $iExitCode ) {
        &logTrace ( 20247,
                  "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                  $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
        my @sys = $lsCmdout[0] =~ /\((.*)\)/;
        my @sOut =split(" ", $lsCmdout[0]);
        if(!@sOut){
          $sOut[1] = 0;
          $sys[0] ="";
        }
        my $iCnt = $sOut[1] + 1;

        $sCmdOpts = "FIRST $iCnt \(";
        foreach my $sNode (split(", ", $sys[0] )){
            chop($sNode);
            $sCmdOpts .= "\\$sNode\\\", ";
        }
        $sCmdOpts .= "\\\"$sSyncSlave\\\"\)";
      }
      else {
        &logError ( 20248,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
         $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout) );
         return 0 # Failure..
      }
    }
    else {
      &logError ( 20249,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      return 0; # Failure..
    }

    $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"ALTER SYSTEM SET synchronous_standby_names = \'\"\'\"\'$sCmdOpts\'\"\'\"\'\"";
  }
  else{
    $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"ALTER SYSTEM RESET synchronous_standby_names;\"";
  }
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20200, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20201,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
      &logTrace ( 20202,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20203,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20204,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
  $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"ALTER SYSTEM RESET synchronous_commit;\"";
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20205, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20206,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
      &logTrace ( 20207,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20208,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20209,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
  $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"SELECT pg_reload_conf();\"";
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20210, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                  $hsPgSQLAttr{'EnvFile'},
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20211,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( 0 == $iExitCode )
    {
      &logTrace ( 20212,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    {
      &logError ( 20213,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20214,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
return 1;
}



sub GetWalSenderPid {
  my ($iTillWhen) = @_;
  my $sMyName = ( caller 0 )[3];
  &logTrace ( 20054, "Subroutine <%s> called with 
  TillWhen               [%s]",
  $sMyName, $iTillWhen); 
  my (@lsCmdout, $iExitCode, $iReturnCode, @iWalSenderPid, $sPidLine) = ();
  $iWalSenderPid[0] = 0;
  my $iTimeout= $iTillWhen - time ;
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  my $sCommandArgs = "-A -t -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"select pid from pg_stat_replication where sync_state=\'\"\'\"\'sync\'\"\'\"\'\""; 
  if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){

    &logTrace ( 20055,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
      if ( 0 == $iExitCode ){
                &logTrace ( 20056,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      if ( (defined $lsCmdout[0]) && ( $lsCmdout[0] =~ /\S/) ) {
        @iWalSenderPid = @lsCmdout;
         &logTrace ( 20057,
                "<%s> Replication active. WalSender Pid is [%s]",
                $sMyName, join(chr 012, @lsCmdout) );
        
      }
      else {
        &logWarn ( 20058,
                "<%s> Replication NOT active.",
                $sMyName );
        return 0;
      }

    }
    else {
      &logError ( 20059,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, @lsCmdout );
      return 0 # Failure..
    }
  }
  else {
    &logError ( 20060,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }  

  my ( $iPid, $sPsCmd, $sPsCmdArgs, $sLang, @lsOut ) = ();


  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';


  $sPsCmd = "/bin/ps";
  $sPsCmdArgs = "wwo uid,pid,ppid,args";

  &logTrace ( 20119, "<%s> PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20120,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  foreach $iPid ( @iWalSenderPid ){
    chomp($iPid);
    #    $sPsCmdArgs .= $iPid;
    #-------------------------------
    # Fire command and get details..
    #-------------------------------
    @lsOut = readpipe ( "$sPsCmd $sPsCmdArgs $iPid" ); # From the command..

    $iExitCode = $? >> 8; # Get exit code..

    #------------------------
    # Restore env variables..
    #------------------------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };

    #------------------
    # Check exit code..
    #------------------
    if ( $iExitCode ) {
      &logError ( 20121,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );
      return 0;
      }

    chomp @lsOut;
    shift @lsOut;
    if ( $lsOut[0] =~ /walsender/ ) {
      &logTrace (20215, "Pid found [%s]", $lsOut[0] );
    }
    else{
      &logError (20216, "Pid [%s] not found", $iPid );
      return 0;
    }
  }
  return join( " ", @iWalSenderPid );
}


=pod

 -------------------------------------------------------------------------
 Function : bool IsSynchronous ( IN Integer TillWhen );

 Description : Checks if the underlying PostgreSQL DB is running 
 in Synchronous mode

 Parameters :
 TillWhen                 Time in EPOCH left for this EP.

 
 Returns : boolean value in the set of {0,1} where
           0 Other
           1 Sync
 -------------------------------------------------------------------------

=cut
sub IsSynchronous {
  my ($iTillWhen) = @_;
  my $sMyName = ( caller 0 )[3];
  my $Mode = ();
  &logTrace ( 20217, "Subroutine <%s> called with
  TillWhen           [%s]",
  $sMyName, $iTillWhen); 
  my (@lsCmdout, $iExitCode, $iReturnCode) = ();
  my $iTimeout= $iTillWhen - time ;
  my $sCommand = "$hsPgSQLAttr{'BaseDir'}/psql";
  my $sCommandArgs = "-A -x -h $hsPgSQLAttr{'HostName'} -p $hsPgSQLAttr{'Port'} -c \"show synchronous_commit;\"";
  
  if ( $iReturnCode = PgSQLRep::ExecLocalCmd ( $hsPgSQLAttr{'PostgreSQLUser'},
                                               $hsPgSQLAttr{'EnvFile'},
                                               $sCommand,
                                               $sCommandArgs,
                                               \$iExitCode,
                                               $iTimeout,
                                               \@lsCmdout) ){

    &logTrace ( 20218,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
      if ( 0 == $iExitCode ){
      &logTrace ( 20219,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      chomp $lsCmdout[0];
      if ( grep(/synchronous_commit\|on/,@lsCmdout) ) {
        &logInfo (20220, "Synchronous commit is set to on.");
        return (1,1);
      }
      else {
        &logInfo (20221, "Synchronous commit is set to off.");
        return (1,0);
      }
    }
    else {
      &logError ( 20222,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, @lsCmdout );
      return 0 # Failure..
    }
  }
  else {
    &logError ( 20223,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }  
return 0;
}

=pod

 -------------------------------------------------------------------------
 Function : String GetOptimalStandby ( IN String ResName );

 Description : Gives Optimal node from list of Slaves for failover

 Parameters :
 ResName                 PgSQLRep Resource Name.

 
 Returns : hostname of Slave with maximum RPO
           or 0 on failure
 -------------------------------------------------------------------------

=cut

sub GetOptimalStandby {

  my ( $sGrpName ) = @_;
  my ( $res, $sType, $sPgSQLResource, $sReplicationModeTuples, @PgSQLRep_SG_Res,
       $sPgSQLRepResource, $sLastPrimary, $sSys, $sEnvFile, @lsCmdout, $sCmd, $iExitCode,
       $sPgSQLState, @sOutput, $sGetLSN, $sOptimalStandby, $sMaxLSN, $iIdx, $sSystemList, $sOptimalStandbyState ) = ();
  
  @PgSQLRep_SG_Res = `$sCmdHagrp -resources $sGrpName`;
  
  foreach $res ( @PgSQLRep_SG_Res ){
    chomp($res);
    $sType=`$sCmdHares -value $res Type`;
    chomp($sType);
    if($sType eq 'PgSQLRep'){
      $sPgSQLRepResource = $res;
    }
  }
  &logTrace ( 20244,"PgSQLRep Resource is [%s]", $sPgSQLRepResource );
  $sCmd = "$sCmdHagrp -value $sGrpName UserStrGlobal ";
  $sLastPrimary = readpipe ( $sCmd ); 
  chomp($sLastPrimary);
  &logTrace ( 20233, "Cmd [%s] output [%s] ",
                      $sCmd, $sLastPrimary );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20234,
          " Cmd [%s] failed with exit code [%s].",
          $sCmd, $iExitCode);
  }
 
  if( !$sLastPrimary  ){
    &logInfo( 20270, "Resource [%s] was not Online Anywhere else previously",
         $sPgSQLRepResource);
    return 1;
  }
   &logInfo( 20240, "Resource [%s] was previously Online on Node [%s]",
                   $sPgSQLRepResource, $sLastPrimary );
    if (PgSQLRep::GetAttrValue ("PgSQLResource",
                              $sPgSQLRepResource,
                              "",
                              \$sPgSQLResource )) {
    &logInfo( 20228, "Value for Attribute PgSQLResource of resource [%s] is [%s]",
         $sPgSQLRepResource, $sPgSQLResource);
  }
  else {
    &logError( 20229, "Failed to obtain attribute value for Attribute PgSQLResource. Returning failure");
    return 0;
  }

  $sCmd = "$sCmdHares -action $sPgSQLRepResource GetAttrValForSys -actionargs $sPgSQLResource State -sys $sLastPrimary 2>&1";
  @sOutput= readpipe ( $sCmd ); 

  &logTrace ( 20233, "Cmd [%s] output [%s] ",
                      $sCmd,   join( "\n", @sOutput ) );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20243, "Failed to obtain attribute value for Attribute State. Returning failure");
    return 0;
  }
  $sPgSQLState = $sOutput[1];
  chomp($sPgSQLState);
  &logInfo( 20242, "Value for Attribute State of resource [%s] for Node [%s] is [%s]",
      $sPgSQLResource, $sLastPrimary, $sPgSQLState);

  if ( $sPgSQLState ne 'FAULTED' ){
    &logInfo( 20271, "Resource [%s] was not faluted on Last Primary node [%s]",
         $sPgSQLResource, $sLastPrimary);
    return 1;
  }

  $sCmd = "$sCmdHares -action $sPgSQLRepResource GetAttrValForSys -actionargs $sPgSQLResource ReplicationModeTuples -sys $sLastPrimary 2>&1";
  @sOutput= readpipe ( $sCmd ); 

  &logTrace ( 20233, "Cmd [%s] output [%s] ",
                      $sCmd,   join( "\n", @sOutput ) );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20231, "Failed to obtain attribute value for Attribute ReplicationModeTuples. Returning failure");
    return 0;
  }
  $sReplicationModeTuples = $sOutput[1];
  chomp($sReplicationModeTuples);
	&logInfo( 20230, "Value for Attribute ReplicationModeTuples of resource [%s] for Node [%s] is [%s]",
	      $sPgSQLResource, $sLastPrimary, $sReplicationModeTuples );

  $sCmd = "$sCmdHagrp -value $sGrpName SystemList ";
  @sOutput = readpipe ( $sCmd ); 
  &logTrace ( 20233, "Cmd [%s] output [%s] ",
                      $sCmd,   join( "\n", @sOutput ) );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20234,
          " Cmd [%s] failed with exit code [%s].",
          $sCmd, $iExitCode);
    return 0;
  }
  $sSystemList = $sOutput[0];
  chomp($sSystemList);
  $sMaxLSN = "0";
  $iIdx = 0;
  foreach my $sItem ( split(' ', $sReplicationModeTuples) ) {
    if( $iIdx % 2 != 0 ){
      $sCmd = "$sCmdHares -action $sPgSQLRepResource GetWALReceiveLSN -sys $sSys 2>&1";
      &logTrace ( 20232,"Executing cmd [%s] to get WAL Log Sequence Number on node [%s]",$sCmd, $sSys );
      
      @sOutput= readpipe ( $sCmd ); 

      &logTrace ( 20233, "Cmd [%s] output [%s] ",
                          $sCmd,   join( "\n", @sOutput ) );
      if ( $iExitCode = $? >> 8 ) {
        &logError ( 20234,
              " Cmd [%s] failed with exit code [%s].",
              $sCmd, $iExitCode);
      }
      else {
        &logTrace ( 20235, "Cmd [%s] succeeded with exit code [%s].", $sCmd,  $iExitCode);
        $sGetLSN = $sOutput[2];
        $sGetLSN =~ s/^\s+|\s+$//g;
        if ( ( $sGetLSN > $sMaxLSN ) || ( ( $sGetLSN == $sMaxLSN ) && ( ( ( $sSystemList =~ $sSys ) || ( $sOptimalStandbyState eq "async" ) ) &&  ($sItem eq "sync") ) ) ) {
          $sMaxLSN = $sGetLSN;
          $sOptimalStandby = $sSys;
          $sOptimalStandbyState = $sItem;
        }
        &logInfo ( 20236, "WAL Log Sequence Number for Node [%s] is [%s] ", $sSys, $sGetLSN );
      }
    }
    else{
      $sSys = $sItem;
    }
    $iIdx++;
  }
  &logInfo ( 20237, "Optimal Standby is [%s] ", $sOptimalStandby );

  $sCmd = "$sCmdHagrp -value $sGrpName ClusterFailOverPolicy ";
  @sOutput = readpipe ( $sCmd ); 
  &logTrace ( 20233, "Cmd [%s] output [%s] ",
                      $sCmd,   join( "\n", @sOutput ) );
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20234,
          " Cmd [%s] failed with exit code [%s].",
          $sCmd, $iExitCode);
    return 0;
  }
  if( $sOutput[0] =~ 'Manual' ){
    &logInfo ( 20272, "PgSQLRep group attribute ClusterFailOverPolicy is set to Manual ");
    if( $sSystemList !~ $sOptimalStandby ){
      &logInfo ( 20273, "Optimal Node [%s] for Failover is not in Local Cluster, Terminating...", $sOptimalStandby );
      &logInfo ( 20274, "You may initiate Online manually on Optimal Node using Command < hagrp -online %s -sys %s > ", $sGrpName, $sOptimalStandby );
      return 0;
    }

  }

  return $sOptimalStandby;

}

1; # So PERL knows its included..
